#include "../config.h"
#include "InConnection.h"
#include "Logging.h"
#include "QtellaSub.h"
#include "Gnutella.h"
#include "Address.h"
#include "UploadManager.h"
#include "DownloadManager.h"
#include "Connections.h"
#include "StringManipulation.h"
#include "Pong.h"
#include "PongCache.h"
#include "WidgetConfig.h"

#include <qtimer.h>
#include <qsocket.h>
#include <qhostaddress.h>
#include <qlistbox.h>
#include <qmessagebox.h>
#include <iostream>
#include <sstream>

#undef _DEBUG

InConnection::InConnection(int socket) : _state(Connected), _timeout(0)
{
	_message_length = 0;
	_v06_ok = false;
	_user_agent = "";
	_socket = new QSocket();
	_socket->setSocket(socket);
	// check whether host is blocked or not
	std::string addr( (const char*)(_socket->peerAddress()).toString() );
	for (unsigned i = 0; i < QtellaSub::getInstance()->_widget_config->ui_listbox_blockedhosts->count(); ++i) {
		std::string s( (const char*)QtellaSub::getInstance()->_widget_config->ui_listbox_blockedhosts->text(i) );
		if( s == addr ) {
			_state = Closed;
			//delete _socket;
			_socket->deleteLater(); // can't delete from a slot connected to its signal
			_socket = NULL;
			return;
		}
	}
	// host is not blocked
	connect( _socket, SIGNAL( readyRead() ), SLOT( slotReadyRead() ) );
	_timeout = new QTimer( this );
	connect( _timeout, SIGNAL(timeout()), SLOT( slotTimeout() ) );
	_timeout->start( 10000 );
}  


InConnection::~InConnection()
{
	if( _socket ) delete _socket;
	if( _timeout ) delete _timeout;
}

void InConnection::slotReadyRead()
{
	qDebug( "%d", _socket->bytesAvailable() );
 	QtellaSub* p = QtellaSub::getInstance();
	int r = 0;
	char buffer[ 1025 ];
	r = _socket->readBlock( buffer, 1024 );
	_rec.append( buffer, r );

	try {
		// check whether it is a message
		// format of messages: MSG\r\nContent-Length: <len>\r\n\r\n<msg>
		if( _rec.substr( 0, 3 ) == "MSG" && _rec.find( "\r\n\r\n" ) == std::string::npos ) {
			return; // waiting for complete header
		} else if( _rec.substr( 0, 3 ) == "MSG" && _rec.find( "\r\n\r\n" ) != std::string::npos ) {
			int err = 1;
			std::string::size_type pos;
			if( ( pos = _rec.find( "Content-Length: " ) ) != std::string::npos ) {
				pos = _rec.find( ":", pos );
				std::string::size_type pos2 = _rec.find( "\r\n", pos );
				if( pos2 != std::string::npos ) {
					pos += 2;
					std::string s = _rec.substr( pos, pos2 - pos );
					if( s.size() > 0 && s.size() < 5 && s.find_first_not_of( "0123456789" ) == std::string::npos ) {
						_rec.erase( 0, _rec.find( "\r\n\r\n" ) + 4 );
						_message_length = atoi( s.c_str() );
						err = 0;
					}
				}
			}
			if( err != 0 )
				throw 0; // not valid
		}

		if( _message_length > 0 ) {
			if( _rec.size() >= _message_length ) {
				if( _rec.size() > _message_length )
					_rec.erase( _message_length - 1 );
				QMessageBox::information( NULL, "Incoming Message", _rec.c_str(), QMessageBox::Information );
				throw 0; // quit
			} else
				return; // need more data
		}
		
		// is it a 0.6 incoming connection (ok message)?
		if ( _v06_ok && _rec.find( "\r\n\r\n" ) != std::string::npos &&
			_rec.substr( 0, p->_gnutella_cid.size() + 4 ) == ( p->_gnutella_cid + "/0.6" ) ) {
			_rec.erase( 0, p->_gnutella_cid.size() + 4 );
			std::string::size_type idx1 = _rec.find_first_not_of (" ", 0);
			std::string status = _rec.substr (idx1, 3);
			if( status != "200" )
				throw 1;
			_timeout->stop();
			disconnect( _socket, 0, 0, 0 );
			Logging::getInstance()->log( Logging::Debug, "New incoming connection. v0.6" );
			std::string ip( _socket->peerAddress().toString().latin1() );
			p->connections->addServant( Address( ip, _socket->port() ), _socket, _user_agent, _v06_header );
			_socket = NULL;
			_state = Closed;
			return;
		}

		// is it a 0.4 incoming connection?
		if( _rec.substr( 0, p->_gnutella_id.size() + 14 ) == ( p->_gnutella_id + " CONNECT/0.4\n\n" ) ) {
			_timeout->stop();
			disconnect( _socket, 0, 0, 0 );
			Logging::getInstance()->log( Logging::Debug, "New incoming connection. v0.4" );
			std::string ip( _socket->peerAddress().toString().latin1() );
			p->connections->addServant( Address( ip, _socket->port() ), _socket );
			_socket = NULL;
			_state = Closed;
			return;
		}

		// is it a 0.6 incoming connection (connect)?
		if( _rec.substr( 0, p->_gnutella_cid.size() + 14 ) == ( p->_gnutella_cid + " CONNECT/0.6\r\n" ) &&
			_rec.find( "\r\n\r\n" ) != std::string::npos ) {
			_v06_header = _rec.substr( 0, _rec.find ( "\r\n\r\n" ) + 4 );
			_rec.erase( 0, _rec.find( "\r\n\r\n" ) + 4 );
			std::string::size_type idx1;
			std::string::size_type idx2;
			if( p->_lan_mode ) {
				idx1 = _v06_header.find( "\r\nLAN:", 0 );
				if( idx1 == std::string::npos )
					throw 2;
				idx1 = _v06_header.find_first_not_of( " ", idx1 + 6 );
				idx2 = _v06_header.find( "\r\n", idx1 + 2 );
				if( _v06_header.substr( idx1, idx2 - idx1 ) != p->_gnutella_id )
					throw 3;
			}
			// get user agent
			idx1 = _v06_header.find( "\r\nUser-Agent:", 0 );
			if( idx1 != std::string::npos ) {
				idx1 = _v06_header.find_first_not_of( " ", idx1 + 13 );
				idx2 = _v06_header.find( "\r\n", idx1 + 2 );
				_user_agent = _v06_header.substr( idx1, idx2 - idx1 );
				if( _user_agent.size() == 0 )
					_user_agent = "?";
			} else 
				_user_agent = "?";
			// answer
			QHostAddress ha = _socket->peerAddress();
			std::string remote_ip = std::string( "Remote-IP: " ) + std::string( ha.toString().latin1() );
			if( ! Connections::getInstance()->incomingSlotsFree() ) {
				std::list<Pong> ponglist;
				PongCache::getInstance()->getPong( ponglist, 10 );
				std::list<Pong>::iterator pos = ponglist.begin();
				std::string s ( p->_gnutella_cid );
				s += std::string( "/0.6 503 BUSY\r\n" );
				if( p->_lan_mode )
					s += "LAN: " + p->_gnutella_id + "\r\n";
				s += "User-Agent: Qtella " + std::string( VERSION ) + ADDITIONAL_VERSION_STR + "\r\n";
				s += remote_ip + "\r\n";
				s += "X-Try: ";
				while( pos != ponglist.end() ) {
					s += ( (*pos)._address ).toString();
					++pos;
					if( pos != ponglist.end() ) 
						s += std::string( "," );
				}
				s += std::string( "\r\n\r\n" );
				_socket->writeBlock( s.data(), s.size() );
				_socket->flush();
				throw 0;
			}
			std::string ok( p->_gnutella_cid );
			ok += "/0.6 200 OK\r\n";
			if( p->_lan_mode )
				ok += "LAN: " + p->_gnutella_id + "\r\n";
			ok += "User-Agent: Qtella " + std::string( VERSION ) + ADDITIONAL_VERSION_STR + "\r\n";
			ok += remote_ip + "\r\n\r\n";
			_socket->writeBlock( ok.data(), ok.size() );
			_socket->flush();
			_v06_ok = true;
			return;
	}

	// is it a GET (file upload) request?
	if( _rec.substr(0, 3) == "GET" ) {
		disconnect( _socket, 0, 0, 0 );
		_timeout->stop();
		Logging::getInstance()->log( Logging::Debug, "New upload request." );
		p->upload_manager->addUpload( _rec, _socket );
		_socket = NULL;
		_state = Closed;
		return;
	}

	// is it a GIV request?
	if( _rec.substr( 0, 3 ) == "GIV" && _rec.find( "\n\n" ) != std::string::npos ) {
		disconnect( _socket, 0, 0, 0 );
		_timeout->stop();
		Logging::getInstance()->log( Logging::Debug, "Got GIV request." );
		p->download_manager->updatePush( _rec, _socket );
		_socket = NULL;
		_state = Closed;
		return;
	}
      
	// only crap thrown at us...
	if( _rec.size () > 2048 )
		throw 0;
	} catch( int e ) {
		_timeout->stop();
		//delete _socket;
		_socket->deleteLater(); // can't delete from a slot connected to its signal
		_socket = NULL;
		_state = Closed;
		return;
	} catch( ... ) {
		std::cerr << "Internal error in Inconnection.cpp." << std::endl;
		_timeout->stop();
		//delete _socket;
		_socket->deleteLater(); // can't delete from a slot connected to its signal
		_socket = NULL;
		_state = Closed;
		return;
	}
}


void InConnection::slotTimeout()
{
	_timeout->stop();
	Logging::getInstance()->log( Logging::Error, "Incoming connection timed out!" );
	if( _socket ) {
		//delete _socket;
		_socket->deleteLater(); // can't delete from a slot connected to its signal
		_socket = NULL;
	}
	_state = Closed;
}


int InConnection::state()
{
	return _state;
}
